==> cpm-1.1.0.tar.gz: cpm-1.1.0/models/plogp.c <==
#include "cpm.h"
#include "../logp_mpi-1.4/logp_mpi.h"
#include <malloc.h> 
#include <memory.h> 
#include <string.h>
#include "mpib_utilities.h"

PLogP_model* PLogP_alloc(int n)
{
	PLogP_model* model = (PLogP_model*)malloc(sizeof(PLogP_model));
	model->predictor.predict_p2p = PLogP_predict_p2p;
	// to be continued: model->predictor.predict_X = PLogP_predict_X;
	model->LogGP_predictor.predict_p2p = LogGP_predict_p2p;
	// to be continued: model->LogGP_predictor.predict_X = LogGP_predict_X;
	model->n = n;
	model->logp = (void**)malloc(sizeof(logp_params*) * n * (n - 1) / 2);
	return model;
}

void PLogP_free(PLogP_model* model)
{
	if (model)
	{
		int i;
		int size;
		for (i = 0, size = model->n * (model->n - 1) / 2; i < size; i++)
			logp_free((logp_params*)model->logp[i]);
		free(model);
	}
}

/*! Gets the size of logp data structure */
int PLogP_size(logp_params* logp)
{
	int size = sizeof(int);// nsizes
	size += sizeof(double);// L
	size += sizeof(logp_size_params) * logp->nsizes;// params
	return size;
}

/*! Packs PLogP model to send via MPI */
void* pack(logp_params* logp)
{
	char* buffer = (char*)malloc(PLogP_size(logp));
	char* pos = buffer;
	memcpy(pos, &logp->nsizes, sizeof(int));
	pos += sizeof(int);
	memcpy(pos, &logp->L, sizeof(double));
	pos += sizeof(double);
	memcpy(pos, logp->params, sizeof(logp_size_params) * logp->nsizes);
	return buffer;
}

/*! Unpacks PLogP model after receive via MPI */
logp_params* unpack (void* buffer)
{
	logp_params* logp = (logp_params*)malloc(sizeof(logp_params));
	logp->descr = (char*)malloc(sizeof(char) * 4);
	strcpy(logp->descr, "CPM");
	char* pos = (char*)buffer;
	memcpy(&logp->nsizes, pos, sizeof(int));
	memcpy(&logp->nsizes_max, pos, sizeof(int));
	pos += sizeof(int);
	memcpy(&logp->L, pos, sizeof(double));
	pos += sizeof(double);
	int size = sizeof(logp_size_params) * logp->nsizes;
	logp->params = (logp_size_params*)malloc(size);
	memcpy(logp->params, pos, size);
	return logp;
}

void PLogP_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	PLogP_model** model_)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute logp parameters for %d processors (must be >= 2)\n", size);
		return;
	}

	int sendcount = size - rank - 1;
	logp_params** logp = (logp_params**)malloc(sizeof(logp_params*) * sendcount);

	MPI_Group group;
	MPI_Comm_group(comm, &group);
	int index = 0;
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;
	while (iter_pairs)
	{
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair)
		{
			if (!parallel)
				MPI_Barrier(comm);
			MPI_Group group_pair;
			MPI_Group_incl(group, 2, iter_pair->values, &group_pair);
			MPI_Comm comm_pair;
			MPI_Comm_create(comm, group_pair, &comm_pair);
			if (comm_pair != MPI_COMM_NULL)
			{
				logp_params* tmp;
				logp_measure(&tmp, "CPM", comm_pair, 0, 1, 0,
					precision.min_reps, precision.max_reps, precision.cl, precision.eps,
					msgset.max_diff, (msgset.min_size > 0) ? msgset.min_size : 1, msgset.max_size, msgset.max_num,
					LOGP_MPI_SEND, LOGP_MPI_RECV, LOGP_BUF_STATIC, LOGP_OPT_GM_DIRECT, LOGP_DIAG_SILENT);
				if (rank == iter_pair->values[0])
					logp[index++] = tmp;
				MPI_Comm_free(&comm_pair);
			}
			MPI_Group_free(&group_pair);
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_free_pairs(pairs);
	MPI_Group_free(&group);

	if (rank == 0)
	{
		PLogP_model* model = PLogP_alloc(size);
		int i, j, index;
		for (i = 0, index = 0; i < sendcount; i++, index++)
			model->logp[i] = logp[i];
		for (i = 1; i < size - 1; i++)
		{
			for (j = i + 1; j < size; j++, index++)
			{
				int size;
				MPI_Recv(&size, 1, MPI_INT, i, 0, comm, MPI_STATUS_IGNORE);
				void* buffer = malloc(size);
				MPI_Recv(buffer, size, MPI_CHAR, i, 0, comm, MPI_STATUS_IGNORE);
				model->logp[index] = unpack(buffer);
				free(buffer);
			}
		}
		*model_ = model;
	}
	else
	{
		int i;
		for (i = 0; i < sendcount; i++)
		{
			int size = PLogP_size(logp[i]);
			MPI_Send(&size, 1, MPI_INT, 0, 0, comm);
			void* buffer = pack(logp[i]);
			MPI_Send(buffer, size, MPI_CHAR, 0, 0, comm);
			free(buffer);
			logp_free(logp[i]);
		}
	}
	free(logp);
}

void PLogP_read(FILE* stream, PLogP_model** _model) {
	int n;
	fscanf(stream, "#Number of nodes: %d\n\n\n", &n);
	PLogP_model* model = *_model = PLogP_alloc(n);
	char* tmpfilename = "logp_mpi.tmp";
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for (j = i + 1; j < model->n; j++, index++) {
			int tmp;
			fscanf(stream, "#%d-%d\n", &tmp, &tmp);
			FILE* tmpfile = fopen(tmpfilename, "w");
			if (tmpfile != NULL) {
				char c;
				while((c = fgetc(stream)) != EOF) {
					fputc(c, tmpfile);
					if (c == '\n') {
						if ((c = fgetc(stream)) == '\n') {
							fgetc(stream);
							break;
						} else
							fputc(c, tmpfile);
					}
				}
				fclose(tmpfile); 
			}
			logp_load((logp_params**)&model->logp[index], tmpfilename);
		}
	}
}

void PLogP_write(FILE* stream, const PLogP_model* model) {
	fprintf(stream, "#Number of nodes: %d\n\n\n", model->n);
	char* tmpfilename = "logp_mpi.tmp";
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for (j = i + 1; j < model->n; j++, index++) {
			fprintf(stream, "#%d-%d\n", i, j);
			logp_save((logp_params*)model->logp[index], tmpfilename);
			FILE* tmpfile = fopen(tmpfilename, "r");
			if (tmpfile != NULL) {
				char c;
				while((c = fgetc(tmpfile)) != EOF)
					fputc(c, stream);
				fclose(tmpfile); 
			}
			fprintf(stream, "\n\n");
		}
	}
}

double PLogP_predict_p2p(void* _this, int i, int j, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	return logp_L(logp) + logp_g(logp, M);
}

double LogGP_predict_p2p(void* _this, int i, int j, int M) {
	//the pointer _this point to LogGP_predictior field for struct PLogP_model
	PLogP_model* model = (PLogP_model*)(_this - sizeof(CPM_predictor));
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	double T = logp_L(logp) + logp_g(logp, 1);
	if (M > 1) {
		int M_max = logp->params[logp->nsizes - 1].size;
		T += logp_g(logp, M_max) * (M - 1) / M_max;
	}
	return T;
}

double PLogP_hpredict_sg_linear(void* _this, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	double L = 0;
	double g = 0;
	int i, j, index;
	for (i = 0; i < model->n - 1; i++) {
		for (j = i + 1; j < model->n; j++, index++) {
			logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, 0, 1)];
			L += logp_L(logp);
			g += logp_g(logp, M);
		}
	}
	return L + g * model->n - 1;
}

double LogGP_hpredict_sg_linear(void* _this, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, 0, 1)];
	double T = logp_L(logp) + logp_g(logp, 1);
	if (M > 0)
		T += logp_g(logp, logp->params[logp->nsizes - 1].size) * (M - 1) * 15 / logp->params[logp->nsizes - 1].size;
	T += logp_g(logp, 1) * 14;
	return T;
}

==> cpm-1.2.0.tar.gz: cpm-1.2.0/models/plogp.c <==
#include "cpm.h"
#include "logp_mpi-1.4/logp_mpi.h"
#include <malloc.h> 
#include <memory.h> 
#include <string.h>
#include "MPIBlib/benchmarks/mpib_utilities.h"

PLogP_model* PLogP_alloc(int n)
{
	PLogP_model* model = (PLogP_model*)malloc(sizeof(PLogP_model));
	model->predictor.predict_p2p = PLogP_predict_p2p;
	// to be continued: model->predictor.predict_X = PLogP_predict_X;
	model->LogGP_predictor.predict_p2p = LogGP_predict_p2p;
	// to be continued: model->LogGP_predictor.predict_X = LogGP_predict_X;
	model->n = n;
	model->logp = (void**)malloc(sizeof(logp_params*) * n * (n - 1) / 2);
	return model;
}

void PLogP_free(PLogP_model* model)
{
	if (model)
	{
		int i;
		int size;
		for (i = 0, size = model->n * (model->n - 1) / 2; i < size; i++)
			logp_free((logp_params*)model->logp[i]);
		free(model);
	}
}

/*! Gets the size of logp data structure */
int PLogP_size(logp_params* logp)
{
	int size = sizeof(int);// nsizes
	size += sizeof(double);// L
	size += sizeof(logp_size_params) * logp->nsizes;// params
	return size;
}

/*! Packs PLogP model to send via MPI */
void* pack(logp_params* logp)
{
	char* buffer = (char*)malloc(PLogP_size(logp));
	char* pos = buffer;
	memcpy(pos, &logp->nsizes, sizeof(int));
	pos += sizeof(int);
	memcpy(pos, &logp->L, sizeof(double));
	pos += sizeof(double);
	memcpy(pos, logp->params, sizeof(logp_size_params) * logp->nsizes);
	return buffer;
}

/*! Unpacks PLogP model after receive via MPI */
logp_params* unpack (void* buffer)
{
	logp_params* logp = (logp_params*)malloc(sizeof(logp_params));
	logp->descr = (char*)malloc(sizeof(char) * 4);
	strcpy(logp->descr, "CPM");
	char* pos = (char*)buffer;
	memcpy(&logp->nsizes, pos, sizeof(int));
	memcpy(&logp->nsizes_max, pos, sizeof(int));
	pos += sizeof(int);
	memcpy(&logp->L, pos, sizeof(double));
	pos += sizeof(double);
	int size = sizeof(logp_size_params) * logp->nsizes;
	logp->params = (logp_size_params*)malloc(size);
	memcpy(logp->params, pos, size);
	return logp;
}

void PLogP_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	PLogP_model** model_)
{
	int rank;
	MPI_Comm_rank(comm, &rank);
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot compute logp parameters for %d processors (must be >= 2)\n", size);
		return;
	}

	int sendcount = size - rank - 1;
	logp_params** logp = (logp_params**)malloc(sizeof(logp_params*) * sendcount);

	MPI_Group group;
	MPI_Comm_group(comm, &group);
	int index = 0;
	MPIB_pairs* pairs = MPIB_build_pairs(size);
	MPIB_pairs* iter_pairs = pairs;
	while (iter_pairs)
	{
		MPI_Barrier(comm);
		MPIB_pair* iter_pair = iter_pairs->list;
		while (iter_pair)
		{
			if (!parallel)
				MPI_Barrier(comm);
			MPI_Group group_pair;
			MPI_Group_incl(group, 2, iter_pair->values, &group_pair);
			MPI_Comm comm_pair;
			MPI_Comm_create(comm, group_pair, &comm_pair);
			if (comm_pair != MPI_COMM_NULL)
			{
				logp_params* tmp;
				logp_measure(&tmp, "CPM", comm_pair, 0, 1, 0,
					precision.min_reps, precision.max_reps, precision.cl, precision.eps,
					msgset.max_diff, (msgset.min_size > 0) ? msgset.min_size : 1, msgset.max_size, msgset.max_num,
					LOGP_MPI_SEND, LOGP_MPI_RECV, LOGP_BUF_STATIC, LOGP_OPT_GM_DIRECT, LOGP_DIAG_SILENT);
				if (rank == iter_pair->values[0])
					logp[index++] = tmp;
				MPI_Comm_free(&comm_pair);
			}
			MPI_Group_free(&group_pair);
			iter_pair = iter_pair->next;
		}
		iter_pairs = iter_pairs->next;
	}
	MPIB_free_pairs(pairs);
	MPI_Group_free(&group);

	if (rank == 0)
	{
		PLogP_model* model = PLogP_alloc(size);
		int i, j, index;
		for (i = 0, index = 0; i < sendcount; i++, index++)
			model->logp[i] = logp[i];
		for (i = 1; i < size - 1; i++)
		{
			for (j = i + 1; j < size; j++, index++)
			{
				int size;
				MPI_Recv(&size, 1, MPI_INT, i, 0, comm, MPI_STATUS_IGNORE);
				void* buffer = malloc(size);
				MPI_Recv(buffer, size, MPI_CHAR, i, 0, comm, MPI_STATUS_IGNORE);
				model->logp[index] = unpack(buffer);
				free(buffer);
			}
		}
		*model_ = model;
	}
	else
	{
		int i;
		for (i = 0; i < sendcount; i++)
		{
			int size = PLogP_size(logp[i]);
			MPI_Send(&size, 1, MPI_INT, 0, 0, comm);
			void* buffer = pack(logp[i]);
			MPI_Send(buffer, size, MPI_CHAR, 0, 0, comm);
			free(buffer);
			logp_free(logp[i]);
		}
	}
	free(logp);
}

void PLogP_read(FILE* stream, PLogP_model** _model) {
	int n;
	fscanf(stream, "#Number of nodes: %d\n\n\n", &n);
	PLogP_model* model = *_model = PLogP_alloc(n);
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for (j = i + 1; j < model->n; j++, index++) {
			int tmp;
			fscanf(stream, "#%d-%d\n", &tmp, &tmp);
			logp_load_((logp_params**)&model->logp[index], stream);
			fscanf(stream, "\n");
		}
	}
}

void PLogP_write(FILE* stream, const PLogP_model* model) {
	fprintf(stream, "#Number of nodes: %d\n\n\n", model->n);
	int i, index;
	for (i = 0, index = 0; i < model->n - 1; i++) {
		int j;
		for (j = i + 1; j < model->n; j++, index++) {
			fprintf(stream, "#%d-%d\n", i, j);
			logp_save_((logp_params*)model->logp[index], stream);
			fprintf(stream, "\n");
		}
	}
}

double PLogP_predict_p2p(void* _this, int i, int j, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	return logp_L(logp) + logp_g(logp, M);
}

double LogGP_predict_p2p(void* _this, int i, int j, int M) {
	//the pointer _this point to LogGP_predictior field for struct PLogP_model
	PLogP_model* model = (PLogP_model*)(_this - sizeof(CPM_predictor));
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, i, j)];
	double T = logp_L(logp) + logp_g(logp, 1);
	if (M > 1) {
		int M_max = logp->params[logp->nsizes - 1].size;
		T += logp_g(logp, M_max) * (M - 1) / M_max;
	}
	return T;
}

double PLogP_hpredict_sg_linear(void* _this, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	double L = 0;
	double g = 0;
	int i, j, index;
	for (i = 0; i < model->n - 1; i++) {
		for (j = i + 1; j < model->n; j++, index++) {
			logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, 0, 1)];
			L += logp_L(logp);
			g += logp_g(logp, M);
		}
	}
	return L + g * model->n - 1;
}

double LogGP_hpredict_sg_linear(void* _this, int M) {
	PLogP_model* model = (PLogP_model*)_this;
	logp_params* logp = (logp_params*)model->logp[MPIB_IJ2INDEX(model->n, 0, 1)];
	double T = logp_L(logp) + logp_g(logp, 1);
	if (M > 0)
		T += logp_g(logp, logp->params[logp->nsizes - 1].size) * (M - 1) * 15 / logp->params[logp->nsizes - 1].size;
	T += logp_g(logp, 1) * 14;
	return T;
}
